/*
Copyright 2022, Savanni D'Gerinel <savanni@luminescent-dreams.com>

This file is part of Savanni's AVR library.

This AVR library is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This AVR library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this AVR library. If not, see <https://www.gnu.org/licenses/>.
*/

#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <dio.h>
#include <display.h>
#include "lantern.h"
#include <rfm.h>
#include <sk9822.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define FPS                 15
#define TICKS_PER_SECOND    7812
#define FPS_TIMEOUT         TICKS_PER_SECOND / FPS
#define PULSE_OFF_COUNT     FPS * 5
#define PULSE_ON_COUNT      5

typedef struct {
    bool frame_timeout;
    bool message_arrived;
} status_flags_t;

status_flags_t status_flags;

ISR(TIMER1_COMPA_vect) {
    status_flags.frame_timeout = true;
}

ISR(INT6_vect) {
    status_flags.message_arrived = true;
}

void setup_fps_timer(void) {
    // WGM = 0100: CTC with OCR1nA top
    // CS = 101: clock / 1024
    TCCR1A = 0;
    TCCR1B = _BV(3) | _BV(2) | _BV(0);

    // Set the top for the counter
    OCR1AH = FPS_TIMEOUT >> 8;
    OCR1AL = FPS_TIMEOUT & 0xff;

    // Enable compare on A match interrupt
    TIMSK1 |= _BV(1);
}

int main(void) {
    EIMSK = 1 << INT6 | 1 << INT2;
    EICRA |= 1 << ISC21 | 1 << ISC20;
    EICRB |= 1 << ISC61 | 1 << ISC60;

    /*
    for (int i = 0; i < 4; i++) {
        status_flags.button_press[i] = false;
    }
    */
    status_flags.frame_timeout = false;

    // Disable unneeded modules
    PRR0 = _BV(7) | _BV(5) | _BV(2);
    PRR1 = _BV(7) | _BV(4) | _BV(3) | _BV(0);

    rfm_t radio = (rfm_t){
        .spi    = {
            .clock          = { .ddr = &DDRB, .port = &PORTB, .pin = &PINB, .addr = 1 },
            .data_out       = { .ddr = &DDRB, .port = &PORTB, .pin = &PINB, .addr = 2 },
            .data_in        = { .ddr = &DDRB, .port = &PORTB, .pin = &PINB, .addr = 3 },
            .chip_select    = { .ddr = &DDRB, .port = &PORTB, .pin = &PINB, .addr = 4 },
        },
        .irq    = { .ddr = &DDRE, .port = &PORTE, .pin = &PINE, .addr = 6 },
        .reset  = { .ddr = &DDRD, .port = &PORTD, .pin = &PIND, .addr = 4 },
    };

    rfm_error_e error;
    rfm_init(&radio, (uint8_t [4]){ 0xde, 0xca, 0xfb, 0xad }, 4, &error);
    rfm_listen(&radio);

    /*
    display_t display = {
        .reg = {
            .output         = { .ddr = &DDRF, .port = &PORTF, .pin = &PINF, .addr = 7 },
            .shift_clock    = { .ddr = &DDRF, .port = &PORTF, .pin = &PINF, .addr = 6 },
            .latch_clock    = { .ddr = &DDRF, .port = &PORTF, .pin = &PINF, .addr = 5 },
        }
    };

    display_init(&display);
    display_write_message(&display, "ready");
    _delay_ms(1000);
    */

    /*
    for (int i = 0; i < 4; i++) {
        dio_set_direction(&buttons[i], LINE_IN);
    }
    */

    sk9822_t lights = {
        .clock_pin = { .ddr = &DDRC, .port = &PORTC, .pin = &PINC, .addr = 7 },
        .data_pin = { .ddr = &DDRD, .port = &PORTD, .pin = &PIND, .addr = 6 }
    };

    rgb_t colors[LIGHT_COUNT];

    lantern_t lantern = lantern_new(lights);
    lantern_set_mode(&lantern, normal);

    lantern_step(&lantern, colors);
    sk9822_send(&lights, colors, LIGHT_COUNT);

    setup_fps_timer();

    set_sleep_mode(SLEEP_MODE_IDLE);
    while(1) {
        if (status_flags.frame_timeout) {
            status_flags.frame_timeout = false;
            lantern_step(&lantern, colors);
            sk9822_send(&lights, colors, LIGHT_COUNT);
        } else if (status_flags.message_arrived) {
            status_flags.message_arrived = false;
            char msg[60];
            uint8_t length;
            interrupt_flags_t flags = rfm_interrupts(&radio);
            while (flags.fifo_not_empty) {
                rfm_receive(&radio, (uint8_t *)msg, &length);
                if (!strncmp(msg, "normal", strlen("normal"))) {
                    lantern_set_mode(&lantern, normal);
                } else if (!strncmp(msg, "spooky", strlen("spooky"))) {
                    lantern_set_mode(&lantern, spooky);
                } else if (!strncmp(msg, "eerie", strlen("eerie"))) {
                    lantern_set_mode(&lantern, eerie);
                }
                flags = rfm_interrupts(&radio);
            }
        }
        /*
        if (status_flags.button_press[0]) {
            status_flags.button_press[0] = false;
            lantern_set_mode(&lantern, normal);
        } else if (status_flags.button_press[1]) {
            status_flags.button_press[1] = false;
            lantern_set_mode(&lantern, spooky);
        } else if (status_flags.button_press[2]) {
            status_flags.button_press[2] = false;
            lantern_set_mode(&lantern, eerie);
        */

        sei();
        sleep_mode();
        cli();
    }

    return 0;
}
